#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2017-2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the fingerd Python 3.x module, which is MIT-licensed.
#******************************************************************************
""" Main script of the module.
	Runs the server depending on the configuration content.
"""

import os.path as _path
import string as _string

from sys import stderr as _stderr
from argparse import ArgumentParser as _ArgumentParser
from configparser import ConfigParser as _ConfigParser

from . import FingerServer as _FingerServer, \
	FingerInterface as _FingerInterface, \
	FingerFiction as _FingerFiction, \
	FingerFictionInterface as _FingerFictionInterface, \
	FingerNativeInterface as _FingerNativeInterface

__ldh__ = _string.ascii_letters + _string.digits + '.-'

def to_this_dir(*path):
	""" Returns the path relative to this directory. """

	return _path.normpath(_path.join(_path.dirname(__file__), *path))

def get_server(config_path):
	""" Get the server with the settings described in the configuration. """

	# Load the configuration.
	c = _ConfigParser()
	c.read(config_path)

	bind  = c['server']['bind']
	host  = c['server']['host']
	iface = c['server']['type']

	if   iface == 'native':
		if not 'FingerNativeInterface' in globals():
			print("error: no native interface exists for your system.",
				file = _stderr)
			exit(1)
		iface = _FingerNativeInterface()
	elif iface == 'actions':
		fic = _FingerFiction(c['actions']['ending'], c['actions']['end'])
		fic.load(c['actions']['path'])

		iface = _FingerFictionInterface(fic)
	else:
		if iface != 'dummy':
			print("warning: unknown interface type, falling back on 'dummy'",
				file = _stderr)
		iface = _FingerInterface()

	server = _FingerServer(bind = bind, host = host, interface = iface)

	return server

def main():
	""" Main function for the module. """

	# Non-interactive command line interface.

	ap = _ArgumentParser()
	ap.add_argument('-c', dest = 'config_path',
		help = 'alternate configuration file path',
		default = _path.normpath(_path.join(_path.dirname(__file__),
			'..', 'fingerd.ini')))
	args = ap.parse_args()

	# Get the server, run it.

	server = get_server(args.config_path)
	server.serve_forever()
	server.shutdown()

if __name__ == '__main__':
	main()

# End of file.
