#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2017-2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the fingerd Python 3.x module, which is MIT-licensed.
#******************************************************************************
""" Definitions for the finger server fiction interface.
	This file contains everything to decode and use the actions file. """

import copy as _copy, re as _re, math as _math
import itertools as _itertools
from datetime import datetime as _dt, timedelta as _td

from ._util import FingerInterface as _FingerInterface, \
	FingerUser as _FingerUser, FingerUserDelta as _FingerUserDelta

__all__ = ["FingerFictionInterface", "FingerFiction"]

_toml = None

# ---
# User for the generated.
# ---

class _FictionalUser(_FingerUser):
	""" User for the fingerd fictional user. """

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.__is_idle = True
		self.__active_since = None

	def __repr__(self):
		p = ('login', 'name', 'line', 'office', 'wd', 'shell',
			'logged_in', 'last_login', 'is_idle')
		p = (f"{x} = {repr(getattr(self, x))}" for x in p \
			if getattr(self, x) is not None)
		return f"{self.__class__.__name__}({', '.join(p)})"

	def __add__(self, delta):
		if isinstance(delta, _FictionalUserDelta):
			if delta.is_idle is not None:
				self.is_idle = delta.is_idle

		super().__add__(delta)
		return self

	@property
	def idle(self):
		""" Idle time (simulated). """

		if not self.logged_in or self.__is_idle:
			return super().idle

		# Generate a number of seconds and return it.

		x = (_dt.now() - self.__active_since).seconds
		s = lambda x: math.sin(x * (pi / 2))
		randsecs = int(abs(s(x) + s(x / 4)))

		return _dt.now() - _td(seconds = randsecs)

	@idle.setter
	def idle(self, value):
		# Does nothing as we manage this time.

		pass

	@property
	def is_idle(self):
		""" Is idle. """

		return self.__is_idle

	@is_idle.setter
	def is_idle(self, value):
		value = bool(value)

		if value and not self.__is_idle:
			_FingerUser.idle = _dt.now()
			self.__active_since = dt.now()

		self.__is_idle = value

	@property
	def logged_in(self):
		""" Is the user still logged in? """

		return super().logged_in

	@logged_in.setter
	def logged_in(self, value):
		was_logged_in = self.logged_in
		_FingerUser.logged_in = value

		if self.logged_in and not was_logged_in:
			self.__is_idle = False
			self.__active_since = _dt.now()

class _FictionalUserDelta(_FingerUserDelta):
	""" User delta for the fingerd fictional user. """

	def __init__(self, *_, is_idle = None, idle = None, **kwargs):
		super().__init__(**kwargs)
		self.is_idle = is_idle

	def __repr__(self):
		p = ('name', 'wd', 'line', 'logged_in', 'is_idle')
		p = (f"{x} = {repr(getattr(self, x))}" for x in p \
			if getattr(self, x) is not None)
		return f"{self.__class__.__name__}({', '.join(p)})"

# ---
# Parse a delta string.
# ---

__delta0_re = _re.compile(r"(-?)(([0-9]+[a-z]+)+)")
__delta1_re = _re.compile(r"([0-9])+([a-z]+)")

def _parse_delta(raw):
	try:
		delta = _td()

		sign, elements, _ = __delta0_re.fullmatch(raw).groups()
		sign = (1, -1)[len(sign)]
		for res in __delta1_re.finditer(elements):
			num, typ = res.groups()
			num = int(num)

			if   typ == 'w':
				delta += _td(weeks = sign * num)
			elif typ in 'jd':
				delta += _td(days = sign * num)
			elif typ == 'h':
				delta += _td(hours = sign * num)
			elif typ == 'm':
				delta += _td(minutes = sign * num)
			elif typ == 's':
				delta += _td(seconds = sign * num)
			else:
				raise Exception

		return delta
	except Exception as e:
		return None

# ---
# Action description.
# ---

class _Action:
	""" Base class for actions in a fiction. """

	def __init__(self, time):
		self.__time = time

	@property
	def time(self):
		""" The time offset to the actions. """

		return self.__time

class _UserCreationAction(_Action):
	""" A user has been created. """

	def __init__(self, offset, user):
		super().__init__(offset)
		self.__user = user

	@property
	def user(self):
		""" The user. """

		return self.__user

	def __repr__(self):
		p = (f"{x} = {repr(getattr(self, x))}" for x in ('time', 'user'))
		return f"{self.__class__.__name__}({', '.join(p)})"

class _UserEditionAction(_Action):
	""" A user status has been changed. """

	def __init__(self, offset, login, delta):
		super().__init__(offset)
		self.__login = login
		self.__delta = delta

	@property
	def login(self):
		""" The user's login. """

		return self.__login

	@property
	def delta(self):
		""" The properties to be modified. """

		return self.__delta

	def __repr__(self):
		p = (f"{x} = {repr(getattr(self, x))}" for x in ('time', 'login',
			'delta'))
		return f"{self.__class__.__name__}({', '.join(p)})"

class _UserDeletionAction(_Action):
	""" A user has been deleted. """

	def __init__(self, offset, login):
		super().__init__(offset)
		self.__login = login

	@property
	def login(self):
		""" The user's login. """

		return self.__login

	def __repr__(self):
		p = (f"{x} = {repr(getattr(self, x))}" for x in ('time', 'login'))
		return f"{self.__class__.__name__}({', '.join(p)})"

# ---
# Finger interface to actions.
# ---

class FingerFiction:
	""" Fiction for the finger daemon, meant to be read by the finger
		fiction interface. """

	def __init__(self, end_type, end_time):
		global _toml

		# Load the required modules.

		if _toml is None:
			try:
				import toml
			except ModuleNotFoundError:
				raise ModuleNotFoundError("'toml' module required") from None

			_toml = toml
			del toml

		# Initialize the properties.

		self.__type = 'interrupt'
		self.__start = _td()
		self.__duration = None
		self.__actions  = []

		if end_type is not None:
			if end_type not in ('loop', 'interrupt', 'stop'):
				raise Exception("end type should be 'loop', " \
					"'interrupt' or 'stop'")
			self.__end_type = end_type
		if end_time is not None:
			self.__end_time = end_time

	def load(self, path):
		""" Load from an `actions.toml` file. """

		actions = []
		end_type = None
		end_time = None

		# Read the document and translate all of the timestamps.

		document = _toml.load(path)
		for key in document.keys():
			time = _parse_delta(key)
			if time is None:
				raise Exception("invalid time format")

			for data in document[key]:
				typ = data['type']

				if   typ in ('loop', 'interrupt', 'stop'):
					# Set the ending type and time.

					if end_time is None or end_time > time:
						end_type = typ
						end_time = time
				elif typ == 'create':
					# User creation.

					user = _FictionalUser(data['login'], data.get('name'))
					if 'shell' in data:
						user.shell = data['shell']
					if 'home' in data:
						user.dir = data['home']

					actions.append(_UserCreationAction(time, user))
				elif typ == 'delete':
					# User deletion.

					login = data['login']

					actions.append(_UserDeletionAction(time, login))
				elif typ in ('login', 'logout', 'cd', 'idle'):
					# User edition.

					login = data['login']

					if   typ == 'login':
						kwargs = {'logged_in': True}
						if 'line' in data:
							kwargs['line'] = data['line']
						if 'office' in data:
							kwargs['office'] = data['office']

						delta = _FictionalUserDelta(**kwargs)
					elif typ == 'logout':
						delta = _FictionalUserDelta(logged_in = False)
					elif typ == 'cd':
						delta = _FictionalUserDelta(wd = data['dir'])
					elif typ == 'idle':
						delta = _FictionalUserDelta(is_idle = True)

					actions.append(_UserEditionAction(time, login, delta))
				else:
					raise Exception("invalid action type")

		# Sort and check the actions.

		actions.sort(key = lambda x: x.time)

		if end_type is None:
			# If no ending was given in the script file, we ought to
			# interrupt 10 seconds after the last action.

			end_type = 'interrupt'
			end_time = actions[-1].time + _td(seconds = 10)
		else:
			# Otherwise, we are removing actions that are after the ending.

			actions = [a for a in actions if a.time <= end_time]

		# FIXME: check that incompatible actions, such as double creation
		#        for a user, doesn't occur.

		self.__type = end_type
		self.__start = actions[0].time if actions else _td()
		self.__duration = end_time
		self.__actions = actions

	def get(self, to = None, since = None):
		""" Get the events. """

		if to is None:
			to = self.__duration
		if since is None:
			since = self.__start

		return [action for action in self.__actions \
			if since <= action.time <= to]

	@property
	def type(self):
		""" Type of action flow, either 'loop', 'interrupt' or 'stop'. """

		return self.__type

	@property
	def duration(self):
		""" Maximum offset. """

		return self.__duration

class FingerFictionInterface(_FingerInterface):
	""" Fiction interface, to repeat a scene which the script
		is written in an actions file. """

	def __init__(self, fiction, start = _dt.now()):
		""" Initialize the interface. """

		# Initialize the object properties.

		if isinstance(fiction, FingerFiction):
			self.__as = _copy.copy(fiction)

		# Initialize the object properties.
		# - `fiction`: the script to follow.
		# - `users`: the users.

		self.__fiction = fiction
		self.__users   = {}
		self.__start   = start
		self.__current = None

	def search_users(self, check):
		""" Look for users according to a check. """

		self.update()
		return [_copy.deepcopy(u) for u in self.__users.values() if check(u)]

	def __get_time(self):
		""" Get the current timedelta using the origin and duration. """

		now = _dt.now()

		if   self.__fiction.type == 'loop':
			# Loop.

			return (now - self.__start) % self.__fiction.duration

		# Either 'stop' or 'interrupt' now.
		# These cases only are distinguishable after the end of the fiction.

		if now > self.__start + self.__fiction.duration:
			if self.__fiction.type == 'stop':
				exit(0)

			return self.__fiction.duration

		# We're within the duration of the fiction, so we just return the
		# offset from the start.

		return now - self.__start

	def update(self):
		""" Update the state according to the script. """

		time = self.__get_time()
		actions = []

		# Check if we need to reset (no users left and apply the actions
		# from before the start).

		if self.__current is None or self.__current > time:
			self.__users = {}
			self.__current = _td(0)

			actions += self.__fiction.get(self.__current)

		# Then get the actions up to the current time.

		actions += self.__fiction.get(time, self.__current)
		self.__current = time

		# And apply the actions.

		for a in actions:
			if   isinstance(a, _UserCreationAction):
				self.__users[a.user.login] = _copy.deepcopy(a.user)
			elif isinstance(a, _UserEditionAction):
				self.__users[a.login] += a.delta
			elif isinstance(a, _UserDeletionAction):
				del self.__users[a.login]

# End of file.
