#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2017-2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the fingerd Python 3.x module, which is MIT-licensed.
#******************************************************************************
""" The main server class for the server, with the related utilities that
	will not be presented to the user, is defined in this file.
"""

import string as _string, socket as _socket, signal as _signal, \
	multiprocessing as _multip, socketserver as _socketserver

from io import TextIOWrapper as _TextIOWrapper

from ._util import FingerInterface as _FingerInterface, \
	FingerFormatter as _FingerFormatter, FingerLogger as _FingerLogger, \
	ConfigurationError as _InvalidConfError, BindError as _InvalidBindError

__all__ = ["FingerServer"]

# ---
# Finger/TCP query decoding.
# ---

class _FingerQuery:
	""" A finger query. Requests information about connected or specific
		users on a remote server.

		There are three types of requests recognized by RFC 1288:
		- {C}  is a request for a list of all online users;
		- {Q1} is a request for a local user;
		- {Q2} is a request for a distant user (with hostname).

		/W means the RUIP (program answering the query) should be more
		verbose (this token can be ignored). """

	# "By default, this program SHOULD filter any unprintable data,
	#  leaving only printable 7-bit characters (ASCII 32 through ASCII 126),
	#  tabs (ASCII 9) and CRLFs."

	allowed_chars = "\t !\"#$%&'()*+,-./:;<=>?@[\\]^_`{|}~" \
		+ _string.ascii_letters + _string.digits

	def __init__(self, raw):
		""" Initialize the query object by decoding the data. """

		# Get a character string out of the query.

		raw = ''.join(c for c in raw if c in self.allowed_chars)
		self.line = raw

		# Get elements.

		self.host = None
		self.username = None
		self.verbose = False
		for element in raw.split():
			if element[0] == '/':
				if 'W' in element[1:]:
					self.verbose = True
				continue
			elif self.username:
				raise Exception
			self.username = element

		if self.username and '@' in self.username:
			self.host, *self.username = self.username.split('@')[::-1]
			self.username = '@'.join(self.username[::-1])

# ---
# TCP request handler.
# ---

class _FingerTCPHandler(_socketserver.StreamRequestHandler):
	""" The TCP request handler for our finger daemon.
		Instantiated once per connection to the server (request).
	"""

	def handle(self):
		""" Request handling. """

		inp = _TextIOWrapper(self.rfile, encoding = 'ascii', errors = 'ignore',
			newline = '\r\n')
		outp = _TextIOWrapper(self.wfile, encoding = 'ascii')

		# Get the server parameters and source data.

		fmt    = self.server.params['formatter']
		iface  = self.server.params['interface']
		host   = self.server.params['host']
		logger = self.server.params['logger']
		src    = self.client_address[0]

		# Gather the request.

		try:
			query = _FingerQuery(inp.readline())
		except:
			logger.bad_query(src)
			ans = fmt.format_query_error(host)
			outp.write(ans)
			return

		ans = fmt.format_header(host, query.line)

		# Gather the content.

		if query.host:
			if query.username:
				logger.transmit(src, query.username, query.host)
			else:
				logger.transmit_list(src, query.host)

			cont = iface.transmit_query(query.host, query.username,
				query.verbose)
			ans += cont
		elif query.username:
			logger.search_users(src, query.username)
			users = iface.search_users(lambda x: query.username in x.login)
			first = True
			if not users:
				ans += fmt.format_no_users_error()
			for user in users:
				if first:
					first = False
				else:
					ans += fmt.format_user_sep()

				ans += fmt.format_user(user, query.verbose)
		else:
			logger.list(src)
			users = iface.search_users(lambda x: x.logged_in)
			ans += fmt.format_list(users, query.verbose)

		# Send the answer.

		ans += fmt.format_footer()
		outp.write(ans)

# ---
# TCP servers.
# ---

class _IPv4TCPServer(_socketserver.ThreadingMixIn, _socketserver.TCPServer):
	""" The IPv4 TCP Server to be used """

	address_family = _socket.AF_INET
	allow_reuse_address = True

	def __init__(self, *args, params = {}, **kwargs):
		self.params = params
		super().__init__(*args, **kwargs)

class _IPv6TCPServer(_socketserver.ThreadingMixIn, _socketserver.TCPServer):
	""" The IPv6 TCP Server to be used """

	address_family = _socket.AF_INET6
	allow_reuse_address = True

	def __init__(self, *args, params = {}, **kwargs):
		self.params = params
		super().__init__(*args, **kwargs)

# ---
# Public server class.
# ---

def _runserver(family, addr, port, params):
	""" Run the server (will be run in another thread). """

	logger = params['logger']

	# Make the server corresponding to the entry.

	if   family == _socket.AF_INET:
		server = _IPv4TCPServer((addr, port), _FingerTCPHandler,
			params = params)
	elif family == _socket.AF_INET6:
		server = _IPv6TCPServer((addr, port), _FingerTCPHandler,
			params = params)

	# Run it.

	logger.start(addr, port)
	try: server.serve_forever()
	except KeyboardInterrupt:
		pass
	logger.stop(addr, port)

_ldh = _string.ascii_letters + _string.digits + '.-'

class FingerServer:
	""" The Finger Server class. """

	def __init__(self, bind = 'localhost:79', host = 'EXAMPLE.ORG',
		interface = _FingerInterface(), formatter = _FingerFormatter(),
		logger = _FingerLogger()):

		# Check the host.

		try:
			host = host.upper()
			assert all(c in _ldh for c in host)
		except:
			raise _InvalidHostError(host)

		# Check the interface and formatter classes.

		if not isinstance(interface, _FingerInterface):
			raise InvalidConfigurationError("please base your interface " \
				"on the base class provided by the fingerd module")
		if not isinstance(formatter, _FingerFormatter):
			raise InvalidConfigurationError("please base your formatter " \
				"on the base class provided by the fingerd module")
		if not isinstance(logger, _FingerLogger):
			raise InvalidConfigurationError("please base your logger " \
				"on the base class provided by the fingerd module")

		# Keep the parameters.

		self._host = host
		self._logger = logger
		self._interface = interface
		self._formatter = formatter

		# Make what is to send to the handlers.

		self._params = {'host': host, 'interface': interface,
			'formatter': formatter, 'logger': logger}

		# Check the addresses.

		if not self._makebinds(bind):
			raise _InvalidBindError()

	def _makebinds(self, addresses):
		""" Create the servers data using the bind values. """

		self._servers = []

		# Extract the addresses as (family, host, port) entries.

		values = set()
		for x in map(lambda x: x.strip(), addresses.split(',')):
			host = ''
			if x[0] == '[':
				# Extract the host using this.

				to = x.find(']')
				if to < 0:
					continue
				host = x[1:to]
				x = x[to + 1:]

				# Extract the port.

				if x in ('', ':'):
					port = 79
				else:
					try:
						assert x[0] == ':'
						port = int(x[1:])
					except:
						continue
			else:
				# Extract the host and the port.

				x = x.split(':')
				if len(x) == 1:
					host = x[0]
				else:
					try:
						port = int(x[-1])
					except:
						continue
					host = ':'.join(x[:-1])

			try:
				entries = _socket.getaddrinfo(host, port,
					proto=_socket.IPPROTO_TCP, type=_socket.SOCK_STREAM)
			except:
				continue

			for ent in entries:
				fam = ent[0]
				if not fam in (_socket.AF_INET, _socket.AF_INET6):
					continue

				ip = ent[4][0]
				#ip = _socket.inet_pton(fam, ip)
				port = ent[4][1]
				values.add((fam, ip, port, None))

		self._servers = list(list(i) for i in values)
		return bool(self._servers)

	def start(self):
		""" Bind and start the underlying servers. """

		for entry in self._servers:
			family, addr, port, _ = entry

			# Check if the thread already exists and runs.

			if entry[3] != None and not entry[3].is_alive():
				continue

			# Start the thread.

			entry[3] = _multip.Process(target = _runserver,
				args = (family, addr, port, self._params))
			entry[3].start()

	def stop(self):
		""" Stop the underlying servers. """

		for entry in self._servers:
			# Check if the thread is still here.

			if entry[3] == None:
				continue
			if not entry[3].is_alive():
				entry[3] = None
				continue

			# Stop it.

			entry[3].join()
			entry[3] = None

	def serve_forever(self):
		""" Serve forever. """

		self.start()
		try:
			while True:
				_signal.pause()
		except KeyboardInterrupt:
			pass
		self.stop()

	def shutdown(self):
		""" Shutdown the server, alias to `.stop()`. """

		self.stop()

# End of file.
