#!/usr/bin/env python3
#******************************************************************************
# Copyright (C) 2017-2018 Thomas "Cakeisalie5" Touhey <thomas@touhey.fr>
# This file is part of the fingerd Python 3.x module, which is MIT-licensed.
#******************************************************************************
""" This file defines the base classes for what is used by the
	finger server, directly or indirectly.
"""

__all__ = ["ConfigurationError", "BindError",
	"FingerUser", "FingerUserDelta",
	"FingerFormatter", "FingerInterface", "FingerLogger"]

import sys as _sys, multiprocessing as _multip
from datetime import datetime as _dt

# ---
# Exceptions.
# ---

class ConfigurationError(Exception):
	""" Generic exception for when an invalid configuration
		exception occurs. """

	pass

class BindError(ConfigurationError):
	""" Exception raised when there were errors in the binds. """

	def __init__(self):
		super().__init__("invalid bind")

# ---
# User-related classes.
# ---

class FingerUser:
	""" User description for the finger interface. """

	def __init__(self, login, name):
		""" Initialize the class out of basic information. """

		self.__login     = None
		self.__name   = ''
		self.__line   = None
		self.__office = None
		self.__wd     = None
		self.__shell  = None

		self.__logged_in  = False
		self.__last_login = None
		self.__idle       = None

		self.login = login
		self.name = name

	def __repr__(self):
		p = ('login', 'name', 'line', 'office', 'wd', 'shell',
			'logged_in', 'last_login', 'idle')
		p = (f"{x} = {repr(getattr(self, x))}" for x in p \
			if getattr(self, x) is not None)
		return f"{self.__class__.__name__}({', '.join(p)})"

	def __add__(self, delta):
		if not isinstance(delta, FingerUserDelta):
			raise NotImplemented

		if delta.logged_in != None:
			self.logged_in = delta.logged_in
		if delta.idle != None:
			self.idle = delta.idle
		if delta.line != None:
			self.line = delta.line
		if delta.name != None:
			self.name = delta.name
		if delta.wd != None:
			self.wd = delta.wd
		if delta.office != None:
			self.office = delta.office

		return self

	@property
	def login(self):
		""" Login name, e.g. 'cake'. """

		return self.__login

	@login.setter
	def login(self, value):
		self.__login = value

	@property
	def logged_in(self):
		""" Is the user still logged in? """

		return self.__logged_in

	@logged_in.setter
	def logged_in(self, value):
		value = bool(value)

		now = _dt.now()
		if value and not self.__logged_in:
			self.__last = now
			if self.wd is None:
				self.wd = self.home
		if not value and self.__logged_in:
			self.__idle = now
			self.line = None
			self.office = None
			self.wd = None
			self.shell = None

		self.__logged_in = value

	@property
	def last_login(self):
		""" Last login date. """

		return self.__last_login

	@property
	def idle(self):
		""" Time since the user isn't doing much. """

		return self.__idle

	@idle.setter
	def idle(self, value):
		self.__idle = value

	@property
	def name(self):
		""" Full user name, e.g. 'Jean Dupont'. """

		return self.__name

	@name.setter
	def name(self, value):
		self.__name = value

	@property
	def line(self):
		""" Serial line, e.g. 'tty7'. """

		return self.__line

	@line.setter
	def line(self, value):
		self.__line = value

	@property
	def office(self):
		""" Office name. """

		return self.__office

	@office.setter
	def office(self, value):
		self.__office = value

	@property
	def wd(self):
		""" Current (working) directory. """

		return self.__wd

	@wd.setter
	def wd(self, value):
		self.__wd = value

	@property
	def shell(self):
		""" Current shell. """

		return self.__shell

	@shell.setter
	def shell(self, value):
		self.__shell = value

class FingerUserDelta:
	""" Differences between two states for a user. """

	def __init__(self, *_, name = None, wd = None, line = None,
		logged_in = None, idle = None, office = None):
		self.logged_in = logged_in
		self.idle      = idle
		self.name      = name
		self.wd        = wd
		self.line      = line
		self.office    = office

	def __repr__(self):
		p = ('name', 'wd', 'line', 'logged_in', 'idle', 'office')
		p = (f"{x} = {repr(getattr(self, x))}" for x in p \
			if getattr(self, x) is not None)
		return f"{self.__class__.__name__}({', '.join(p)})"

# ---
# Formatter base class.
# ---

class FingerFormatter:
	""" Answer formatter.
		Formats the answer following RFC 1288. """

	def format_query_error(self, name):
		""" There is an error in your query! """

		return "Site: {}\r\n" \
			"You have made a mistake in your query!\r\n".format(name)

	def format_header(self, name, request):
		""" Formats the header of each request. """

		if request:
			request = ' ' + request

		return f"Site: {name}\r\n" f"Command line:{request}\r\n" "\r\n"

	def format_footer(self):
		""" Formats the footer of each request (in case the developer
			wants to add something at the end, anything). """

		return ""

	def format_user(self, user, verbose = False):
		""" Formats user information. """

		return 'Login name: {:<27} In real life: {}\r\n'.format(user.login,
			user.name)

	def format_user_sep(self):
		""" Formats the user separator. """

		return '\r\n'

	def format_list(self, users, verbose = False):
		""" Format a user list. """

		if not users:
			return 'No user list available.\r\n'

		now = _dt.now()

		lt = 'Login Time'
		columns = (
			('Login',)  + tuple(u.login for u in users),
			('Name',)   + tuple(u.name for u in users),
			('TTY',)    + tuple(u.line if u.line else "" for u in users),
			('Idle',)   + tuple("=TODO=" for u in users),
			(lt,)       + tuple("=TODO=" for u in users),
			('Office',) + tuple(u.office if u.office else '' for u in users))
		columns_min = (8, 14, 6, 5, 12, 10)

		sizes = tuple(max(max(map(len, c)), columns_min[i]) + 1 \
			for i, c in enumerate(columns))

		lines = []
		for line in range(1 + len(users)):
			lines.append(' '.join('{:<{}}'.format(columns[i][line], sizes[i]) \
				for i in range(len(columns))))

		return '\r\n'.join(lines) + '\r\n'

	def format_no_users_error(self):
		""" No users correspond to the search! """

		return 'No users correspond to your query!\r\n'

# ---
# Interface (dummy) base class.
# ---

class FingerInterface:
	""" Finger interface to the users.
		Expandable (the current class defines the dummy interface). """

	def transmit_query(self, user, host, verbose):
		return "This server won't transmit finger queries.\r\n"

	def search_users(self, check):
		return []

# ---
# Server logger base class.
# ---

class FingerLogger:
	""" Finger logger for the server.
		Expandable, but the base class should be enough for most needs. """

	def __init__(self, stream = _sys.stderr):
		self._stream = stream
		self._lock = _multip.Lock()

	def _log(self, fmt, *args):
		self._lock.acquire()
		print('\r' + fmt.format(*args), file=self._stream)
		self._lock.release()

	def _logr(self, src, fmt, *args):
		self._log('[{}] ' + fmt, src, *args)

	def start(self, host, port):
		self._log("Starting fingerd on [{}]:{}.", host, port)

	def stop(self, host, port):
		self._log("Stopping fingerd on [{}]:{}.", host, port)

	def bad_query(self, source):
		self._logr(source, "bad request.")

	def transmit_list(self, source, host):
		self._logr(source, "transmit user list query to `{}`.", host)

	def transmit(self, source, username, host):
		self._logr(source, "transmit user query for `{}` to `{}`.",
			username, host)

	def search_users(self, source, username):
		self._logr(source, "look for user `{}`.", username)

	def list(self, source):
		self._logr(source, "list connected users.")

# End of file.
